package hashedstore;
@ISA = qw(Exporter);
@EXPORT = qw(store_data delete_file_and_link);
use warnings;
use strict;
#use IO::Handle;
use POSIX;
use Digest::MD5;
use Errno;

our $debug = 0;

sub store_data ($$;$)
{
   my $filename = shift;
   my $data = shift;
   my $mode = shift; # || 0444);
   my $changed = 0; # 1 = file, 2 = symlink

   unless ($filename =~ m:^(.*?)([^/]*)$:) {
      die "could not match filename = <$filename>\n";
   }
   my ($dir, $fil) = ($1, $2);
   if ($dir) {
      die "$dir is not a directory" unless -d $dir;
   }
   die "empty or illegal filename <$filename>" if $fil eq '' or $fil eq '.' 
      or $fil eq '..';

   my $digest = Digest::MD5::md5_base64($data);
   $digest =~ tr/\//%/;
   $debug && print "filename = <$filename>, digest = <$digest>\n";

   my $digbase = $fil . '#' . $digest;
   my $digfile = $dir . $digbase;
#
#  try to create file if it does not exist
#
   my $rc;
   if ($mode) {
      $rc = sysopen (F, $digfile, O_RDWR | O_CREAT | O_EXCL, $mode);
   }
   else {
      $rc = sysopen (F, $digfile, O_WRONLY | O_CREAT | O_EXCL);
   }
   if (not defined $rc) {
      if (exists &Errno::EEXIST) {
         if ($!{EEXIST} != $!) {
            die "sysopen: $!";
         }
      }
      else {
         warn "sysopen: $!";
      }
   }
   else {
      my ($cl, $wl) = length $data;
      $wl = syswrite (F, $data) or $! and die "syswrite: $!";
      warn "cl = <$cl>, wl = <$wl>\n" if $cl ne $wl;
      close (F) or die "close: $!";
      $changed |= 1;
   }
#
#  check if link points to this file
#
   my $olddirfile = readlink $filename;
   if (not defined $olddirfile) { # no link
      if (exists &Errno::ENOENT) {
         if ($!{ENOENT} != $!) {
            die "readlink <$filename>: $!";
         }
      }
      else {
         warn "readlink: $!";
      }
      warn "link <$filename> has gone, re-creating ...\n"
         if $changed == 0;
   }
   elsif ($olddirfile ne $digbase) {
      warn "link <$filename> has wrong target <$olddirfile>, unlinking and correcting ...\n"
         if $changed == 0;
#      $olddirfile = undef if not -e $olddirfile;
   }
   else {
      return $changed; # link already okay
   }
#
#  create (new) symlink (atomically)
#
   symlink $digbase, $filename . '_' or die "link: $!";
   if (! rename ($filename . '_', $filename)) {
      unlink $filename; # or warn "unlink <$filename>: $!";
      die "could not unlink $filename" if -l $filename;
      rename ($filename . '_', $filename) or die "rename: $!";
   }
   $changed |= 2;
   if (unlink ($filename . '_')) {
      warn "deleted $filename"."_ after successful rename\n";
   }; 
#   printf "olddirfile = <$olddirfile>\n";
   if (defined $olddirfile) {
      if ($olddirfile =~ /$fil#[0-9a-zA-Z+%]{22}/) {
         $rc = unlink ($dir . $olddirfile); # returns the number of files deleted
         if ($rc == 0) {
            if (exists &Errno::ENOENT) {
               if ($!{ENOENT} != $!) {
                  die "unlink $dir . $olddirfile: $!";
               }
            }
            warn "unlink " . $dir . $olddirfile . ": $!";
         }
      }
      else {
         warn "keeping old link target <$olddirfile>\n";
      }
   }
   return $changed; # file or link changed
}

sub delete_file_and_link ($)
{
   my $filename = shift;
   my $deleted = 0;
   unless ($filename =~ m:^(.*?)([^/]*)$:) {
      die "could not match filename = <$filename>\n";
   }
   my ($dir, $fil) = ($1, $2);

   my $olddirfile = readlink $filename;
   if (not defined $olddirfile) { # no link
      if (exists &Errno::ENOENT) {
         if ($!{ENOENT} != $!) {
            warn "readlink <$filename>: $!";
            return undef;
         }
      }
      else {
         warn "readlink <$filename>: $!";
      }
      warn "no such link <$filename>\n";
      return undef;
   }
   if ($olddirfile !~ /$fil#[0-9a-zA-Z+%]{22}/) {
      warn "keeping link <$filename> and target <$olddirfile>: form mismatch\n";
      return undef;
   }
   my $rc = unlink ($dir . $olddirfile);
   if ($rc == 0) {
      if (exists &Errno::ENOENT) {
         if ($!{ENOENT} != $!) {
            die "unlink $dir . $olddirfile: $!";
         }
      }
      warn "unlink " . $dir . $olddirfile . ": $!";
   }
   $deleted |= 1;
   $rc = unlink ($filename);
   if ($rc == 0) {
      if (exists &Errno::ENOENT) {
         if ($!{ENOENT} != $!) {
            die "unlink <$filename>: $!";
         }
      }
      warn "unlink <$filename>: $!";
   }
   $deleted |= 1;
   return $deleted;
}

1;
